
/* globals.c -- written by Istvan Varga, 2003 */

#define CS_H_NO_MACROS_FOR_GLOBALS  1

#include <cs.h>
#include "oload.h"

/* IV - Jan 30 2003: to have access to the real exit() function */
#ifdef exit
#undef exit
#endif

/* default Csound API functions */

static const CSOUNDAPI default_api = {
    /* general functions */
    csoundGetVersion,
    csoundGetAPIVersion,
    csoundGetHostData,
    csoundSetHostData,
    csoundPerform,
    csoundCompile,
    csoundPerformKsmps,
    csoundPerformBuffer,
    csoundCleanup,
    csoundReset,
    /* attributes */
    csoundGetGlobals,
    csoundGetSr,
    csoundGetKr,
    csoundGetGlobalKr,
    csoundGetKsmps,
    csoundGetGlobalKsmps,
    csoundGetNchnls,
    csoundGet0dBFS,
    csoundGetSampleFormat,
    csoundGetSampleSize,
    csoundGetInputBufferSize,
    csoundGetOutputBufferSize,
    csoundGetInputBuffer,
    csoundGetOutputBuffer,
    csoundGetSpin,
    csoundGetSpout,
    csoundGetScoreTime,
    csoundGetProgress,
    csoundGetProfile,
    csoundGetCpuUsage,
    /* score */
    csoundIsScorePending,
    csoundSetScorePending,
    csoundGetScoreOffsetSeconds,
    csoundSetScoreOffsetSeconds,
    csoundRewindScore,
    /* messages */
    csoundMessage,
    csoundMessageV,
    csoundMessageS,
    csoundThrowMessage,
    csoundThrowMessageV,
    csoundSetMessageCallback,
    csoundSetMessageVCallback,
    csoundSetMessageSCallback,
    csoundSetThrowMessageCallback,
    csoundGetMessageLevel,
    csoundSetMessageLevel,
    csoundInputMessage,
    /* control */
    csoundKeyPress,
    csoundGetChar,
    csoundSetInputValueCallback,
    csoundInputValue,
    csoundSetOutputValueCallback,
    csoundOutputValue,
    csoundScoreEvent,
    /* MIDI */
    csoundSetExternalMidiOpenCallback,
    csoundSetExternalMidiReadCallback,
    csoundSetExternalMidiWriteCallback,
    csoundSetExternalMidiCloseCallback,
    csoundIsExternalMidiEnabled,
    csoundSetExternalMidiEnabled,
    /* display/graphics */
    csoundSetIsGraphable,
    csoundSetMakeGraphCallback,
    csoundSetDrawGraphCallback,
    csoundSetKillGraphCallback,
    csoundSetExitGraphCallback,
    /* opcodes */
    csoundNewOpcodeList,
    csoundDisposeOpcodeList,
    csoundAppendOpcode,
    /* libraries */
    csoundLoadExternal,
    csoundLoadExternals,
    csoundOpenLibrary,
    csoundCloseLibrary,
    csoundGetLibrarySymbol,
    /* misc. functions */
    csoundSetYieldCallback,
    csoundSetEnv,
    csoundGetEnv,
    /* real-time audio */
    csoundSetPlayopenCallback,
    csoundSetRtplayCallback,
    csoundSetRecopenCallback,
    csoundSetRtrecordCallback,
    csoundSetRtcloseCallback,
    /* error handling */
    csoundSetInitErrorCallback,
    csoundInitError,
    csoundSetPerfErrorCallback,
    csoundPerfError,
    csoundSetDieCallback,
    csoundDie,
    csoundSetErrorHandlerCallback,
    csoundError,
    /* opcode helper functions */
    csoundGetEngineForOpcode,
    csoundGetGlobalsForOpcode,
    csoundGetOpcodeParams,
    csoundGetMIDIParams,
    csoundQueryFTableInit,
    csoundQueryFTablePerf,
    csoundQueryDeferredFTableInit,
    csoundGetReleaseTime,
    csoundSetReleaseTime,
    csoundGetReleaseFlag,
    csoundGetBeatModeFlag,
    csoundGetOffTime,
    csoundSetOffTime,
    csoundGetPFields,
    csoundSkipInitOpcodes,
    csoundSkipPerfOpcodes,
    csoundTurnOffCurrentNote,
    csoundAuxAlloc,
    csoundOpenFile,
    csoundCloseFile,
    csoundLoadFile,
    csoundRandomValueU,
    csoundRandomValueB,
    csoundSetRandomSeed,
    /* time functions */
    timers_struct_init,
    timers_get_real_time,
    timers_get_CPU_time,
    timers_random_seed,
    /* memory management */
    csoundHeapCreate,
    csoundHeapSetErrorCode,
    csoundMalloc,
    csoundCalloc,
    csoundFree,
    csoundRealloc,
    csoundHeapDestroy,
    /* native language support */
    csoundLocalizeString
};

/* -- standard Csound globals that can be accessed from user applications -- */

static const CSGLOBALS default_csglobals = {
    DFLT_KSMPS,                     /* ksmps of current instrument instance  */
    DFLT_KSMPS,                     /* orchestra level ksmps                 */
    (MYFLT) DFLT_KSMPS,             /* ksmps as float/double                 */
    (MYFLT) DFLT_KSMPS,             /* global_ksmps as float/double          */
    DFLT_KR,                        /* control rate of curr. instr. instance */
    DFLT_KR,                        /* orchestra level control rate          */
    (FL(1.0) / DFLT_KR),            /* FL(1.0) / ekr                         */
    (FL(1.0) / DFLT_KR),            /* FL(1.0) / global_ekr                  */
    (FL(0.5) / DFLT_KR),            /* FL(0.5) / ekr                         */
    (FL(0.5) / DFLT_KR),            /* FL(0.5) / global_ekr                  */
    (FMAXLEN / DFLT_KR),            /* FMAXLEN / ekr                         */
    (FMAXLEN / DFLT_KR),            /* FMAXLEN / global_ekr                  */
    0L,                             /* number of k-cycles in onedkr units    */
    0L,                             /* nr of k-cycles in global_onedkr units */
    DFLT_SR,                        /* sample rate                           */
    (FL(1.0) / DFLT_SR),            /* FL(1.0) / esr                         */
    (FMAXLEN / DFLT_SR),            /* FMAXLEN / esr                         */
    (TWOPI_F / DFLT_SR),            /* 2PI / esr                             */
    (-PI_F / DFLT_SR),              /* -PI / esr                             */
    (-TWOPI_F / DFLT_SR),           /* -2PI / esr                            */
    DFLT_DBFS,                      /* 0dBFS level (32767 by default)        */
    FL(0.0),                        /* unused                                */
    DFLT_NCHNLS,                    /* number of audio channels              */
    0,                              /* global_ksmps * nchnls                 */
    0,                              /* global_ksmps * nchnls                 */
    0,                              /* non-zero if already wrote to spout    */
    NULL,                           /* float audio in buffer (nspin samples) */
    NULL,                           /* float audio out buffer (nspout smpls) */
    -1L,                            /* highest k-rate ZAK location           */
    -1L,                            /* highest a-rate ZAK location           */
    NULL,                           /* k-rate ZAK space (NULL if none)       */
    NULL,                           /* a-rate ZAK space (NULL if none)       */
    NULL,                           /* curr. instr. instance (i-time only !) */
    NULL,                           /* current init opcode (i-time only)     */
    NULL,                           /* current perf opcode (perf time only)  */
    NULL,                           /* pointer to message level (-m) value   */
    0,                              /* non-zero in reinit mode               */
    0,                              /* non-zero for tied notes               */
    NULL,                           /* pointer to API functions              */
    NULL,                           /* pointer to array of strset strings    */
    0,                              /* maximum strset array index            */
    0                               /* unused                                */
};

/* should be removed at some time when support for multiple */
/* instances is fully implemented */
GLOBALS     cglob;

/* void globalsRESET(void *csound)                              */
/*   reset all global variables of a Csound instance, and       */
/*   install default CsoundAPI callback functions on first call */

#define GLOBALS_MAGIC_1 (0x75E63A4D)
#define GLOBALS_MAGIC_2 (0x1B9AF4CB)

static int globals_are_initialised(GLOBALS *csound)
{
    CSOUNDAPI *cs_api = (CSOUNDAPI*) csound;

    if (cs_api->GetVersion == (int (*)(void)) csoundGetVersion &&
        csound->csMagic1 == GLOBALS_MAGIC_1 &&
        csound->csMagic2 == GLOBALS_MAGIC_2)
      return 1;
    else
      return 0;
}

void globalsRESET(void *csound)
{
    CSOUNDAPI   saved_api;
    GLOBALS     *cs_globals = (GLOBALS*) csound;
    CSOUNDAPI   *cs_api = (CSOUNDAPI*) csound, *api_ptr;
    int         keep_api, i;

    keep_api = globals_are_initialised(cs_globals);
    /* if not first initialisation, save API function pointers */
    if (keep_api) {
      api_ptr = &saved_api;
      memcpy(api_ptr, cs_api, sizeof(CSOUNDAPI));
    }
    else {
      /* otherwise use default interface */
      api_ptr = (CSOUNDAPI*) &default_api;
    }

    /* clear entire structure to zero */
    memset(cs_globals, 0, sizeof(GLOBALS));
    /* set up API function pointers */
    memcpy(cs_api, api_ptr, sizeof(CSOUNDAPI));
    /* OPARMS structure */
    cs_globals->O.sfwrite     = 1;
    cs_globals->O.sfheader    = 1;
    cs_globals->O.displays    = 1;
#ifdef LINUX
    cs_globals->O.msglevel    = 231;
#elif defined(WIN32)
    cs_globals->O.msglevel    = 135;
#else
    cs_globals->O.msglevel    = 7;
#endif
    /* standard globals */
    cs_globals->csglobals = default_csglobals;
    cs_globals->csglobals.msglevel = &(cs_globals->O.msglevel);
    cs_globals->csglobals.csound = cs_api;
    /* mark as initialised */
    cs_globals->csMagic1 = GLOBALS_MAGIC_1;
    cs_globals->csMagic2 = GLOBALS_MAGIC_2;

    /* set any remaining variables that do not have a */
    /*   zero or NULL default value                   */

    /* orchestra header variables calculated by otran */
    cs_globals->tran_sr       = FL(-1.0);
    cs_globals->tran_kr       = FL(-1.0);
    cs_globals->tran_ksmps    = FL(-1.0);
    cs_globals->tran_0dbfs    = DFLT_DBFS;
    cs_globals->tran_nchnls   = DFLT_NCHNLS;
    /* 0dBFS system */
    cs_globals->dbfs_to_short = FL(32767.0) / DFLT_DBFS;
    cs_globals->short_to_dbfs = DFLT_DBFS / FL(32767.0);
    cs_globals->dbfs_to_float = FL(1.0) / DFLT_DBFS;
    cs_globals->float_to_dbfs = DFLT_DBFS;
    cs_globals->dbfs_to_long  = FL(2147483647.0) / DFLT_DBFS;
    cs_globals->long_to_dbfs  = DFLT_DBFS / FL(2147483647.0);
    /* real-time audio device numbers for Win32 */
    cs_globals->rtin_dev      = 1024u;
    cs_globals->rtout_dev     = 1024u;
    /* non-zero if nchnls > 1 */
    cs_globals->multichan     = (DFLT_NCHNLS > 1 ? 1 : 0);
    /* oload.h (???) */
    cs_globals->nlabels       = NLABELS;
    cs_globals->ngotos        = NGOTOS;
    cs_globals->peakchunks    = 1;
    /* misc. global variables (FIXME: to be documented ... ) */
    cs_globals->holdrand      = (int32_t) 2345678;
    cs_globals->maxinsno      = MAXINSNO;
    cs_globals->maxopcno      = -1;                 /* IV - Oct 24 2002 */
    cs_globals->Linefd        = -1;
    cs_globals->midi_out      = -1;
    cs_globals->oparms        = &(cs_globals->O);

    /* set up a default exitjmp handler */
    /* user code can override it for re-entrancy */
    if ((i = setjmp(cs_globals->exitjmp)) != 0) {
      if (i == CSOUND_EXITJMP_NOERROR)
        exit(0);
      else
        exit(i);
    }
}

/* same as globalsRESET(), but has no effect */
/* if globals are already initialised */

void globals_init(void *csound)
{
    if (!globals_are_initialised((GLOBALS*) csound))
      globalsRESET(csound);
}

