
/* ------ oscils, lphasor, and tablexkt by Istvan Varga (Jan 5 2002) ------ */

#ifndef CSOUND_OSCILS_H
#define CSOUND_OSCILS_H

/* oscils opcode struct */

typedef struct {
    OPDS    h;
    MYFLT   *ar, *iamp, *icps, *iphs, *iflg;                /* opcode args  */
    /* internal variables */
    int     use_double;
    double  xd, cd, vd;
    MYFLT   x, c, v;
} OSCILS;

/* lphasor opcode struct */

typedef struct {
    OPDS    h;
    MYFLT   *ar, *xtrns, *ilps, *ilpe;                      /* opcode       */
    MYFLT   *imode, *istrt, *istor;                         /* args         */
    /* internal variables */
    int     loop_mode;
    double  phs, lps, lpe;
    int     dir;            /* playback direction (0: backward, 1: forward) */
} LPHASOR;

/* tablexkt opcode struct */

typedef struct {
    OPDS    h;
    MYFLT   *ar, *xndx, *kfn, *kwarp, *iwsize;              /* opcode       */
    MYFLT   *ixmode, *ixoff, *iwrap;                        /* args         */
    /* internal variables */
    int     raw_ndx, ndx_scl, wrap_ndx, wsize;
    MYFLT   win_fact;
/*  double  wsized2_d, pidwsize_d; */           /* for oscils_hann.c */
} TABLEXKT;

typedef struct {                /* denorm a1[, a2[, a3[, ... ]]] */
    OPDS    h;
    MYFLT   *ar[256];
} DENORMS;

typedef struct {                /* kr delayk ksig, idel[, imode] */
    OPDS    h;
    MYFLT   *ar, *ksig, *idel, *imode;
    int     npts, init_k, readp, mode;
    AUXCH   aux;
} DELAYK;

typedef struct {                /* kr vdel_k ksig, kdel, imdel[, imode] */
    OPDS    h;
    MYFLT   *ar, *ksig, *kdel, *imdel, *imode;
    int     npts, init_k, wrtp, mode;
    MYFLT   frstkval;
    AUXCH   aux;
} VDELAYK;

typedef struct {                /* ar rbjeq asig, kfco, klvl, kQ, kS[, imode] */
        OPDS    h;
        MYFLT   *ar, *asig, *kcps, *klvl, *kQ, *kS, *imode;     /* args */
        /* internal variables */
        MYFLT   old_kcps, old_klvl, old_kQ, old_kS;
        double  omega, cs, sn;
        MYFLT   xnm1, xnm2, ynm1, ynm2;
        MYFLT   b0, b1, b2, a1, a2;
        int     ftype;
} RBJEQ;

/* these functions are exported to entry*.c */

#ifndef CSOUND_OSCILS_C
extern void oscils_set(void*), oscils(void*);
extern void lphasor_set(void*), lphasor(void*);
extern void tablexkt_set(void*), tablexkt(void*);

extern void denorms(void*);
extern void delaykset(void*), delayk(void*);
extern void vdelaykset(void*), vdelayk(void*);

extern void rbjeqset(void*);
extern void rbjeq(void*);
#endif

#endif              /* CSOUND_OSCILS_H */

