
#include <Csound/csdl.h>
#include "soundout.h"

/* IV - Feb 18 2003: rewritten sndout opcodes */

/* functions from soundin.c */

static void bytrev2(char *buf, int nbytes)  /* reverse bytes in buf of shorts */
{
    uint8_t *p = (uint8_t*) buf, c;
    int     n = nbytes >> 1;

    do {
      n--;
      c = p[0];
      p[0] = p[1];
      p[1] = c;
      p += 2;
    } while (n);
}

static void bytrev4(char *buf, int nbytes)  /* reverse bytes in buf of longs */
{
    uint8_t *p = (uint8_t*) buf, c;
    int     n = nbytes >> 2;

    do {
      n--;
      c = p[0];
      p[0] = p[3];
      p[3] = c;
      c = p[1];
      p[1] = p[2];
      p[2] = c;
      p += 4;
    } while (n);
}

/* from soundio.c */

static void sndwrterr(SNDOUT *p, unsigned nret, unsigned nput)
/* report soundfile write error         */
/* called after chk of write() bytecnt  */
{
    ENGINE->Message(ENGINE,
                    Str("soundfile write returned bytecount of %u, not %u\n"),
                    nret, nput);
    ENGINE->Message(ENGINE,
                    Str("(disk may be full...\n closing the file ...)\n"));
    ENGINE->Die(ENGINE, Str("\t... closed\n"));
}

/* RWD:DBFS: NB: thse funcs all supposed to write to a 'raw' file, so
   what will people want for 0dbfs handling? really need to update
   opcode with more options. */

#define SNDOUT_CLIP(x,min,max)                                          \
    ((x) < (min) ? (min) : ((x) > (max) ? (max) : (x)))

/* IV - Feb 18 2003: 8 bit signed */

static void sndwrt1(SNDOUT *p, int fd, MYFLT *buf, int nsamps)
{
    int     n;
    int8_t  *buffer = (int8_t*) p->c.outbuf2;
    MYFLT   sclval = FL(127.5) / GLOBLS->e0dbfs;

    for (n = 0; n < SNDOUTSMPS; n++) {
      long x = (long) (buf[n] * sclval);
      x = SNDOUT_CLIP(x, -128L, 127L);
      buffer[n] = (int8_t) x;
    }
    if ((n = write(fd, (void*) buffer, nsamps)) < nsamps)
      sndwrterr(p, n, nsamps);
}

/* IV - Feb 18 2003: 8 bit unsigned */

static void sndwrtu(SNDOUT *p, int fd, MYFLT *buf, int nsamps)
{
    int     n;
    uint8_t *buffer = (uint8_t*) p->c.outbuf2;
    MYFLT   sclval = FL(127.5) / GLOBLS->e0dbfs;

    for (n = 0; n < SNDOUTSMPS; n++) {
      long x = (long) (buf[n] * sclval);
      x = SNDOUT_CLIP(x, -128L, 127L);
      buffer[n] = (uint8_t) (x ^ 0x80L);
    }
    if ((n = write(fd, (void*) buffer, nsamps)) < nsamps)
      sndwrterr(p, n, nsamps);
}

/* IV - Feb 18 2003: 16 bit inverse byte order */

static void sndwrt2rev(SNDOUT *p, int fd, MYFLT *buf, int nsamps)
{
    int     n, nbytes;
    int16_t *buffer = (int16_t*) p->c.outbuf2;
    MYFLT   sclval = FL(32767.0) / GLOBLS->e0dbfs;

    for (n = 0; n < SNDOUTSMPS; n++) {
      long x = (long) (buf[n] * sclval);
      buffer[n] = (int16_t) SNDOUT_CLIP(x, -32768L, 32767L);
    }
    bytrev2((char*) buffer, nbytes = (nsamps << 1));  /* rev bytes in shorts */
    if ((n = write(fd, (void*) buffer, nbytes)) < nbytes)
      sndwrterr(p, n, nbytes);
}

/* IV - Feb 18 2003: 16 bit native byte order */

static void sndwrt2(SNDOUT *p, int fd, MYFLT *buf, int nsamps)
{
    int     n, nbytes;
    int16_t *buffer = (int16_t*) p->c.outbuf2;
    MYFLT   sclval = FL(32767.0) / GLOBLS->e0dbfs;

    for (n = 0; n < SNDOUTSMPS; n++) {
      long x = (long) (buf[n] * sclval);
      buffer[n] = (int16_t) SNDOUT_CLIP(x, -32768L, 32767L);
    }
    nbytes = nsamps << 1;
    if ((n = write(fd, (void*) buffer, nbytes)) < nbytes)
      sndwrterr(p, n, nbytes);
}

/* IV - Feb 18 2003: 24 bit little-endian */

static void sndwrt3le(SNDOUT *p, int fd,MYFLT *buf,int nsamps)
{
    int     n, nbytes;
    uint8_t *buffer = (uint8_t*) p->c.outbuf2, *bufp;
    long    x;
    MYFLT   sclval = FL(8388607.0) / GLOBLS->e0dbfs;

    for (n = 0, bufp = buffer; n < SNDOUTSMPS; n++) {
      x = (long) (buf[n] * sclval);
      x = SNDOUT_CLIP(x, -8388608L, 8388607L);
      *(bufp++) = (uint8_t) (x & 0x000000FFL); x >>= 8;
      *(bufp++) = (uint8_t) (x & 0x000000FFL); x >>= 8;
      *(bufp++) = (uint8_t) (x & 0x000000FFL);
    }
    nbytes = nsamps * 3;
    if ((n = write(fd, (void*) buffer, nbytes)) < nbytes)
      sndwrterr(p, n, nbytes);
}

/* IV - Feb 18 2003: 24 bit big-endian */

static void sndwrt3be(SNDOUT *p, int fd,MYFLT *buf,int nsamps)
{
    int     n, nbytes;
    uint8_t *buffer = (uint8_t*) p->c.outbuf2, *bufp;
    long    x;
    MYFLT   sclval = FL(8388607.0) / GLOBLS->e0dbfs;

    for (n = 0, bufp = buffer; n < SNDOUTSMPS; n++) {
      x = (long) (buf[n] * sclval);
      x = SNDOUT_CLIP(x, -8388608L, 8388607L);
      *(bufp++) = (uint8_t) ((x & 0x00FF0000L) >> 16);
      *(bufp++) = (uint8_t) ((x & 0x0000FF00L) >> 8);
      *(bufp++) = (uint8_t)  (x & 0x000000FFL);
    }
    nbytes = nsamps * 3;
    if ((n = write(fd, (void*) buffer, nbytes)) < nbytes)
      sndwrterr(p, n, nbytes);
}

/* IV - Feb 18 2003: 32 bit inverse byte order */

static void sndwrt4rev(SNDOUT *p, int fd, MYFLT *buf, int nsamps)
{
    int     n, nbytes;
    int32_t *buffer = (int32_t*) p->c.outbuf2;
    double  x;
    MYFLT   sclval = FL(2147483647.0) / GLOBLS->e0dbfs;

    for (n = 0; n < SNDOUTSMPS; n++) {
      x = (double) (buf[n] * sclval);
      x = SNDOUT_CLIP(x, -2147483648.25, 2147483647.25);
      buffer[n] = (int32_t) x;
    }
    nbytes = nsamps << 2;
    bytrev4((char*) buffer, nbytes);    /* rev bytes in longs   */
    if ((n = write(fd, (void*) buffer, nbytes)) < nbytes)
      sndwrterr(p, n, nbytes);
}

/* IV - Feb 18 2003: 32 bit native byte order */

static void sndwrt4(SNDOUT *p, int fd, MYFLT *buf, int nsamps)
{
    int     n, nbytes;
    int32_t *buffer = (int32_t*) p->c.outbuf2;
    double  x;
    MYFLT   sclval = FL(2147483647.0) / GLOBLS->e0dbfs;

    for (n = 0; n < SNDOUTSMPS; n++) {
      x = (double) (buf[n] * sclval);
      x = SNDOUT_CLIP(x, -2147483648.25, 2147483647.25);
      buffer[n] = (int32_t) x;
    }
    nbytes = nsamps << 2;
    if ((n = write(fd, (void*) buffer, nbytes)) < nbytes)
      sndwrterr(p, n, nbytes);
}

/* IV - Feb 18 2003: float native byte order */

static void sndwrtf(SNDOUT *p, int fd, MYFLT *buf, int nsamps)
{
    int     n, nbytes;
#ifdef MYFLT_IS_FLOAT
    nbytes = nsamps << 2;
    if ((n = write(fd, (void*) buf, nbytes)) < nbytes)
      sndwrterr(p, n, nbytes);
#else
    float   *buffer = (float*) p->c.outbuf2;
    /* This code is only needed in MYFLT == double case */
    for (n = 0; n < SNDOUTSMPS; n++) {
      buffer[n] = (float) buf[n];
    }
    nbytes = nsamps << 2;
    if ((n = write(fd, (void*) buffer, nbytes)) < nbytes)
      sndwrterr(p, n, nbytes);
#endif
}

/* IV - Feb 18 2003: float inverse byte order */

static void sndwrtfrev(SNDOUT *p, int fd, MYFLT *buf, int nsamps)
{
    int     n, nbytes;
#ifdef MYFLT_IS_FLOAT
    nbytes = nsamps << 2;
    bytrev4((char*) buf, nbytes);
    if ((n = write(fd, (void*) buf, nbytes)) < nbytes)
      sndwrterr(p, n, nbytes);
#else
    float   *buffer = (float*) p->c.outbuf2;
    /* This code is only needed in MYFLT == double case */
    for (n = 0; n < SNDOUTSMPS; n++) {
      buffer[n] = (float) buf[n];
    }
    nbytes = nsamps << 2;
    bytrev4((char*) buffer, nbytes);
    if ((n = write(fd, (void*) buffer, nbytes)) < nbytes)
      sndwrterr(p, n, nbytes);
#endif
}

static void sndoset_(SNDOUT *p)         /* init routine for instr soundout   */
{
    int    soutfd, endian_reqd, endian_host = 0x1234;
    char   *sfname, sndoutname[256], err_msg[256];
    SNDCOM *c = &(p->c);

    if (c->fdch.fd != 0)
      return;                           /* if file already open, rtn  */
    if (*c->ifilcod == SSTRCOD) {
      int   outocount, xoutcode, inocount, xincode;
      char  **strargs;
      /* if named file: */
      ENGINE->GetOpcodeParams(p, &outocount, &xoutcode, &inocount, &xincode,
                                 &strargs);
      /* get string argument */
      strcpy(sndoutname, strargs[0]);
    }
    else {
      int   filno = (int) (*c->ifilcod + FL(0.25));
      /* numbered file */
      if (GLOBLS->strsets != NULL && filno <= GLOBLS->strsmax &&
          GLOBLS->strsets[filno] != NULL)
        /* use strset string if available */
        strcpy(sndoutname, GLOBLS->strsets[filno]);
      else
        sprintf(sndoutname, "soundout.%d", (int) (*c->ifilcod + FL(0.25)));
    }
    sfname = sndoutname;
    if ((soutfd = ENGINE->OpenFile(p, &(c->fdch), sfname, 1)) < 0) {
      /* if openout failed */
      sprintf(err_msg, Str("soundout cannot open %s"), sfname);
      /* else just print the errmsg */
      ENGINE->InitError(ENGINE, err_msg);
      return;
    }

    if (c->format > 0) {
      c->format |= 0x100;
      /* requested endianness: */
      /* 0x00: native / default, 0x10: inverse, 0x20: little, 0x30: big */
      endian_reqd = (c->format & 0x30) >> 4;
      c->format &= 0x1CF;
    }
    else endian_reqd = 0;
    /* host endianness 0: little 1: big */
    endian_host = (*((uint8_t*) (&endian_host)) == (uint8_t) 0x34 ? 0 : 1);

    if (c->format == AE_24INT) {
      /* for 24 bit ints: 0: little-endian, 1: big-endian */
      endian_reqd = (endian_reqd & 02 ? (endian_reqd & 01)
                                        : ((endian_reqd & 01) ^ endian_host));
    }
    else {
      /* other formats: 0: native, 1: reverse */
      endian_reqd = (endian_reqd & 02 ? ((endian_reqd & 01) ^ endian_host)
                                        : (endian_reqd & 01));
    }
    switch (c->format) {
      case AE_UNCH:
        c->swrtmethod = sndwrtu;                                break;
      case AE_CHAR:
        c->swrtmethod = sndwrt1;                                break;
      case AE_SHORT:
        c->swrtmethod = (endian_reqd ? sndwrt2rev : sndwrt2);   break;
      case AE_24INT:
        c->swrtmethod = (endian_reqd ? sndwrt3be : sndwrt3le);  break;
      case AE_LONG:
        c->swrtmethod = (endian_reqd ? sndwrt4rev : sndwrt4);   break;
      case AE_FLOAT:
        c->swrtmethod = (endian_reqd ? sndwrtfrev : sndwrtf);   break;
      default:
        /* IV - Feb 18 2003: might as well trap a bad format param */
        ENGINE->InitError(ENGINE,
                          Str("Error: bad format parameter for soundout"));
        return;
    }
    if (c->format == AE_24INT)
      endian_reqd ^= endian_host;
    if (ENGINE->GetMessageLevel(ENGINE)) {
      ENGINE->Message(ENGINE,
                      "soundout: opening outfile %s (RAW), with%s bytrev\n",
                      sfname, endian_reqd ? "" : " no");
    }
    c->outbufp = (MYFLT*) c->outbuf;        /* fix - isro 20-11-96 */
    c->bufend = c->outbufp + SNDOUTSMPS;    /* fix - isro 20-11-96 */
}

static void soundout(SNDOUT *p)
{
    MYFLT  *bufp, *endp, **asigs;
    int    nn = 0, chn, nsmps, nchns;

    if (p->c.fdch.fd < 1) {
      ENGINE->PerfError(ENGINE, Str("soundout: not initialised"));
      return;
    }
    asigs = p->c.args;
    bufp = p->c.outbufp;
    endp = p->c.bufend;
    nsmps = GLOBLS->ksmps;
    nchns = p->c.nchns;
    do {
      chn = 0;
      do {
        *(bufp++) = asigs[chn][nn];
        if (bufp >= endp) {
          bufp = (MYFLT*) p->c.outbuf;
          p->c.swrtmethod(p, p->c.fdch.fd, (MYFLT*) p->c.outbuf,
                          (int) (endp - bufp));
        }
      } while (++chn < nchns);
    } while (++nn < nsmps);
    /* save buffer position */
    p->c.outbufp = bufp;
}

/* deactivation routine */

static void soundout_d(SNDOUT *p)
{
    MYFLT  *bufp, *endp;

    if (p->c.fdch.fd < 1)       /* if the instrument was deactivated       */
      return;                   /* because of an error, return immediately */
    bufp = (MYFLT*) p->c.outbuf;
    endp = p->c.outbufp;
    /* flush output buffer */
    if ((endp - bufp) > 0)
      p->c.swrtmethod(p, p->c.fdch.fd, (MYFLT*) bufp, (int) (endp - bufp));
}

static void sndo1set(SNDOUT *p)
{
    p->c.format = (int) (*(p->iformat) + FL(0.25));
    p->c.nchns = 1;
    p->c.p = (SNDOUT*) p;
    p->c.ifilcod = p->ifilcod;
    p->c.args = (MYFLT**) &(p->asig);
    sndoset_(p);
}

static void sndo2set(SNDOUTS *p)
{
    p->c.format = (int) (*(p->iformat) + FL(0.25));
    p->c.nchns = 2;
    p->c.p = (SNDOUT*) p;
    p->c.ifilcod = p->ifilcod;
    p->c.args = (MYFLT**) &(p->asig1);
    sndoset_((SNDOUT*) p);
}

static void sndocset(SNDOUTC *p)
{
    p->c.format = (int) (*(p->iformat) + FL(0.25));
    {
      int   outocount, xoutcode, inocount, xincode;
      char  **strargs;
      ENGINE->GetOpcodeParams(p, &outocount, &xoutcode, &inocount, &xincode,
                                 &strargs);
      p->c.nchns = inocount - 2;
    }
    if (p->c.nchns < 1 || p->c.nchns > 24) {
      ENGINE->InitError(ENGINE, Str("soundoutc: invalid number of arguments"));
      return;
    }
    p->c.p = (SNDOUT*) p;
    p->c.ifilcod = p->ifilcod;
    p->c.args = (MYFLT**) (p->asigs);
    sndoset_((SNDOUT*) p);
}

void OPCODE(soundout_opcode_init) (void* csound)
{
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "soundout",   sizeof(SNDOUT),     0x15,   "",     "aSo",
      (SUBR) sndo1set,  (SUBR) NULL,    (SUBR) soundout,    (SUBR) soundout_d);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "soundouts",  sizeof(SNDOUTS),    0x15,   "",     "aaSo",
      (SUBR) sndo2set,  (SUBR) NULL,    (SUBR) soundout,    (SUBR) soundout_d);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "soundoutc",  sizeof(SNDOUTC),    0x15,   "",     "Siy",
      (SUBR) sndocset,  (SUBR) NULL,    (SUBR) soundout,    (SUBR) soundout_d);
}

